"""Client-side UDP connection."""

from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator

from aioudp import connection


class _ClientProtocol(asyncio.DatagramProtocol):
    def __init__(
        self,
        on_connection: asyncio.Future[connection.Connection],
        on_connection_lost: asyncio.Future[bool],
        queue_size: int | None = None,
    ) -> None:
        self.on_connection = on_connection
        self.on_connection_lost = on_connection_lost
        self.msg_queue: asyncio.Queue[bytes | None] = (
            asyncio.Queue() if queue_size is None else asyncio.Queue(queue_size)
        )

    def connection_made(
        self,
        transport: asyncio.DatagramTransport,  # type: ignore[override]
        # I am aware of the Liskov subsitution principle
        # but asyncio.DatagramProtocol had this function signature
    ) -> None:
        self.on_connection.set_result(
            connection.Connection(
                send_func=transport.sendto,
                recv_func=self.msg_queue.get,
                is_closing=transport.is_closing,
                get_local_addr=lambda: transport.get_extra_info("sockname"),
                get_remote_addr=lambda: transport.get_extra_info("peername"),
            ),
        )

    def datagram_received(self, data: bytes, _addr: connection.AddrType) -> None:
        self.msg_queue.put_nowait(data)

    def error_received(self, exc: BaseException) -> None:
        raise exc

    def connection_lost(self, exc: Exception | None) -> None:
        self.msg_queue.put_nowait(None)
        self.on_connection_lost.set_result(True)
        if exc:
            raise exc


@asynccontextmanager
async def connect(
    host: str,
    port: int,
    queue_size: int | None = None,
    **kwargs: Any,  # noqa: ANN401
) -> AsyncIterator[connection.Connection]:
    """Connect to a UDP server.

    See Also:
        :func:`serve`

    Args:
        host (str): The server's host name/address.
        port (int): The server's port number.
        queue_size (int | None):
            The maximum size of the message queue used internally.
            Defaults to None, meaning an unlimited size. Unless you know for sure
            what you're doing, there is no need to change this value.
        **kwargs:
            Additional keyword arguments to pass to

            `loop.create_datagram_endpoint <https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.create_datagram_endpoint>`_.


    Returns:
        An asynchronous iterator yielding a connection to the UDP server.

    """
    loop = asyncio.get_running_loop()
    on_connection = loop.create_future()
    on_connection_lost = loop.create_future()
    transport, _ = await loop.create_datagram_endpoint(
        lambda: _ClientProtocol(on_connection, on_connection_lost, queue_size),
        remote_addr=(host, port),
        **kwargs,
    )

    conn = await on_connection
    try:
        yield conn
    finally:
        transport.close()
        await on_connection_lost



"""Connection class for aioudp."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Awaitable, Callable, Tuple

from aioudp import exceptions

__all__ = ["Connection"]

AddrType = Tuple[str, int]


@dataclass
class Connection:  # TODO(ThatXliner): REFACTOR: minimal args
    # https://github.com/ThatXliner/aioudp/issues/15
    """Represents a server-client connection. Do not instantiate manually."""

    send_func: Callable[[bytes], None]
    recv_func: Callable[[], Awaitable[bytes | None]]
    is_closing: Callable[[], bool]
    get_local_addr: Callable[[], AddrType]
    get_remote_addr: Callable[[], None | AddrType]
    closed: bool = False

    @property
    def local_address(self) -> AddrType:
        """Returns the local address of the connection. This is your IP.

        .. seealso::
            :meth:`remote_address`

        Returns:
            tuple[str, int]: This is a `tuple` containing the hostname and port

        """
        return self.get_local_addr()

    @property
    def remote_address(self) -> None | AddrType:
        """Returns the remote address of the connection. This is their IP.

        .. seealso::
            :meth:`local_address`

        Returns:
            tuple[str, int]: This is a `tuple` containing the hostname and port

        """
        return self.get_remote_addr()

    async def recv(self) -> bytes:
        """Receives a message from the connection.

        Returns:
            bytes: The received `bytes`.

        Raises:
            exceptions.ConnectionClosedError: The connection is closed

        """
        output = await self.recv_func()
        if output is None:
            raise exceptions.ConnectionClosedError
        return output

    async def send(self, data: bytes) -> None:
        """Send a message to the connection.

        The reason why this send function is `async` is due to API consistency.
        There is actually no underlying `async` call so feel free to
        just forgo the `await`.

        .. warning::
            Since this is UDP, there is no guarantee that the message will be sent

        Args:
            data (bytes): The message in bytes to send

        Raises:
            exceptions.ConnectionClosedError: The connection is closed
            ValueError: There is no data to send

        """
        if self.is_closing():
            msg = "The connection is closed"
            raise exceptions.ConnectionClosedError(msg)
        if not data:
            msg = "You must send some data"
            raise ValueError(msg)
        self.send_func(data)

    def __aiter__(self) -> Connection:
        """Return an async iterator of messages received."""
        return self

    async def __anext__(self) -> bytes:
        """Get the next message received. Functionally equivalent to :meth:`recv`."""
        try:
            return await self.recv()
        except exceptions.ConnectionClosedError as error:
            raise StopAsyncIteration from error



"""Server-side UDP connection."""

from __future__ import annotations

import asyncio
import functools
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import Any, AsyncIterator, Callable, Coroutine, NoReturn

from aioudp import connection


@dataclass
class _ServerProtocol(asyncio.DatagramProtocol):
    handler: Callable[[connection.Connection], Coroutine[Any, Any, None]]
    queue_size: int | None
    msg_queues: dict[connection.AddrType, asyncio.Queue[bytes | None]] = field(
        default_factory=dict,
    )
    msg_handlers: dict[
        connection.AddrType,
        asyncio.Task[None],
    ] = field(
        default_factory=dict,
    )
    transport: None | asyncio.transports.DatagramTransport = None

    def connection_made(
        self,
        transport: asyncio.DatagramTransport,  # type: ignore[override]
        # I am aware of the Liskov subsitution principle
        # but asyncio.DatagramProtocol had this function signature
    ) -> None:
        self.transport = transport

    def datagram_received(self, data: bytes, addr: connection.AddrType) -> None:
        if addr not in self.msg_queues:
            self.msg_queues[addr] = (
                asyncio.Queue()
                if self.queue_size is None
                else asyncio.Queue(self.queue_size)
            )
            assert self.transport is not None

            def done(_) -> None:  # type: ignore[no-untyped-def]  # noqa: ANN001
                del self.msg_queues[addr]
                del self.msg_handlers[addr]

            # Strong reference is needed
            self.msg_handlers[addr] = asyncio.create_task(
                self.handler(  # See connnection.py
                    connection.Connection(  # TODO(ThatXliner): REFACTOR: minimal args
                        # https://github.com/ThatXliner/aioudp/issues/15
                        send_func=functools.partial(self.transport.sendto, addr=addr),
                        recv_func=self.msg_queues[addr].get,
                        is_closing=self.transport.is_closing,
                        get_local_addr=functools.partial(
                            self.transport.get_extra_info,
                            "sockname",
                        ),
                        # This should theoretically be
                        # the same as the `addr` parameter
                        get_remote_addr=functools.partial(
                            self.transport.get_extra_info,
                            "peername",
                        ),
                    ),
                ),
            )
            self.msg_handlers[addr].add_done_callback(done)

        self.msg_queues[addr].put_nowait(data)

    def error_received(self, exc: Exception) -> NoReturn:
        # Haven't figured out why this can happen
        raise exc

    # The server is done
    def connection_lost(self, exc: None | Exception) -> None:
        for key in self.msg_queues:
            self.msg_queues[key].put_nowait(None)
            self.msg_handlers[key].cancel()
        # Haven't figured out why this can happen
        if exc is not None:
            raise exc


@asynccontextmanager
async def serve(
    host: str,
    port: int,
    handler: Callable[[connection.Connection], Coroutine[Any, Any, None]],
    queue_size: int | None = None,
    **kwargs: Any,  # noqa: ANN401
) -> AsyncIterator[None]:
    """Run a UDP server.

    See the docs for an example UDP echo server

    See Also:
        :func:`connect`

        :doc:`An example UDP echo server </index>`

    Args:
        host (str): The host name/address to run the server on
        port (int): The port number to run the server on
        handler (Callable[[connection.Connection], Awaitable[None]]):
            An asynchronous function to handle a request
            It should accept an instance of :class:`connection.Connection`
            and doesn't need to return anything.

        queue_size (int | None):
            The maximum size of the message queue used internally.
            Defaults to None, meaning an unlimited size. Unless you know for sure
            what you're doing, there is no need to change this value.
        **kwargs:
            Additional keyword arguments to pass to

            `loop.create_datagram_endpoint <https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.create_datagram_endpoint>`_.

    """
    loop = asyncio.get_running_loop()
    transport: asyncio.BaseTransport
    _: asyncio.BaseProtocol
    transport, _ = await loop.create_datagram_endpoint(
        lambda: _ServerProtocol(handler, queue_size),
        local_addr=(host, port),
        **kwargs,
    )
    try:
        yield
    finally:
        transport.close()
